"""
Very simple transformations that qiskit may even fail or hard to control
"""

from typing import Any, Dict, List, Optional, Tuple, Union
from copy import copy

import numpy as np

from ..abstractcircuit import AbstractCircuit
from ..cons import backend
from ..quantum import QuOperator
from .. import gates
from ..utils import is_sequence


def replace_r(circuit: AbstractCircuit, **kws: Any) -> AbstractCircuit:
    qir = circuit.to_qir()
    c: Any = type(circuit)(**circuit.circuit_param)
    for d in qir:
        if "parameters" not in d:
            c.apply_general_gate_delayed(d["gatef"], d["name"], mpo=d["mpo"])(
                c, *d["index"], split=d["split"]
            )
        else:
            if d["gatef"].n == "rx":
                c.h(*d["index"])
                c.rz(*d["index"], theta=d["parameters"].get("theta", 0.0))
                c.h(*d["index"])

            elif d["gatef"].n == "ry":
                c.sd(*d["index"])
                c.h(*d["index"])
                c.rz(*d["index"], theta=d["parameters"].get("theta", 0.0))
                c.h(*d["index"])
                c.s(*d["index"])

            elif d["gatef"].n == "rzz":
                c.cx(*d["index"])
                c.rz(d["index"][1], theta=d["parameters"].get("theta", 0.0))
                c.cx(*d["index"])

            elif d["gatef"].n == "rxx":
                c.h(d["index"][0])
                c.h(d["index"][1])
                c.cx(*d["index"])
                c.rz(d["index"][1], theta=d["parameters"].get("theta", 0.0))
                c.cx(*d["index"])
                c.h(d["index"][0])
                c.h(d["index"][1])

            elif d["gatef"].n == "ryy":
                c.sd(d["index"][0])
                c.sd(d["index"][1])
                c.h(d["index"][0])
                c.h(d["index"][1])
                c.cx(*d["index"])
                c.rz(d["index"][1], theta=d["parameters"].get("theta", 0.0))
                c.cx(*d["index"])
                c.h(d["index"][0])
                c.h(d["index"][1])
                c.s(d["index"][0])
                c.s(d["index"][1])

            else:
                c.apply_general_variable_gate_delayed(
                    d["gatef"], d["name"], mpo=d["mpo"]
                )(c, *d["index"], **d["parameters"], split=d["split"])

    return c  # type: ignore


def replace_u(circuit: AbstractCircuit, **kws: Any) -> AbstractCircuit:
    qir = circuit.to_qir()
    c: Any = type(circuit)(**circuit.circuit_param)
    for d in qir:
        if "parameters" not in d:
            c.apply_general_gate_delayed(d["gatef"], d["name"], mpo=d["mpo"])(
                c, *d["index"], split=d["split"]
            )
        else:
            if d["gatef"].n == "u":
                c.rz(*d["index"], theta=d["parameters"].get("lbd", 0) - np.pi / 2)
                c.h(*d["index"])
                c.rz(*d["index"], theta=d["parameters"].get("theta", 0))
                c.h(*d["index"])
                c.rz(*d["index"], theta=np.pi / 2 + d["parameters"].get("phi", 0))
            else:
                c.apply_general_variable_gate_delayed(
                    d["gatef"], d["name"], mpo=d["mpo"]
                )(c, *d["index"], **d["parameters"], split=d["split"])

    return c  # type: ignore


def _get_matrix(qir_item: Dict[str, Any]) -> Any:
    if "gate" in qir_item:
        op = qir_item["gate"]
    else:
        op = qir_item["gatef"](**qir_item["parameters"])
    if isinstance(op, QuOperator):
        m = backend.numpy(op.eval_matrix())
    else:
        m = backend.numpy(backend.reshapem(op.tensor))
    return m


def prune(
    circuit: Union[AbstractCircuit, List[Dict[str, Any]]],
    rtol: float = 1e-3,
    atol: float = 1e-3,
    **kws: Any
) -> Any:
    if isinstance(circuit, list):
        qir = circuit
        output = "qir"
    else:
        qir = circuit.to_qir()
        output = "tc"
    if output in ["tc", "circuit"]:
        c: Any = type(circuit)(**circuit.circuit_param)  # type: ignore
        for d in qir:
            m = _get_matrix(d)

            if not np.allclose(
                m / (m[0, 0] + 1e-8), np.eye(m.shape[0]), rtol=rtol, atol=atol
            ):
                # upto a phase
                if "parameters" not in d:
                    c.apply_general_gate_delayed(d["gatef"], d["name"], mpo=d["mpo"])(
                        c, *d["index"], split=d["split"]
                    )
                else:
                    c.apply_general_variable_gate_delayed(
                        d["gatef"], d["name"], mpo=d["mpo"]
                    )(c, *d["index"], **d["parameters"], split=d["split"])
        return c

    elif output in ["qir"]:
        nqir = []
        for d in qir:
            m = _get_matrix(d)

            if not np.allclose(
                m / (m[0, 0] + 1e-8), np.eye(m.shape[0]), rtol=rtol, atol=atol
            ):
                nqir.append(d)
                # upto a phase

        return nqir


# upto global phase
default_merge_rules = {
    ("s", "s"): "z",
    ("sd", "sd"): "z",
    ("t", "t"): "s",
    ("td", "td"): "sd",
    ("x", "y"): "z",
    ("y", "x"): "z",
    ("x", "z"): "y",
    ("z", "x"): "y",
    ("z", "y"): "x",
    ("y", "z"): "x",
    ("x", "x"): "i",
    ("y", "y"): "i",
    ("z", "z"): "i",
    ("h", "h"): "i",
    ("rz", "rz"): "rz",
    ("rx", "rx"): "rx",
    ("ry", "ry"): "rx",
    ("rzz", "rzz"): "rzz",
    ("rxx", "rxx"): "rxx",
    ("ryy", "ryy"): "ryy",
    ("crz", "crz"): "crz",
    ("crx", "crx"): "crx",
    ("cry", "cry"): "crx",
    ("cnot", "cnot"): "i",
    ("cz", "cz"): "i",
    ("cy", "cy"): "i",
}


def _find_next(qir: List[Dict[str, Any]], i: int) -> Optional[int]:
    # if qir[i] is None:
    # return None
    index = qir[i]["index"]
    for j, item in enumerate(qir[i + 1 :]):
        # if item is not None:
        if item["index"] == index:
            return j + i + 1
        for ind in item["index"]:
            if ind in index:
                return None
    return None


def _get_theta(qir_item: Dict[str, Any]) -> float:
    theta = qir_item["parameters"].get("theta", 0.0)
    if is_sequence(theta) and len(theta) == 1:
        return theta[0]  # type: ignore
    return theta  # type: ignore


def _merge(
    qir: List[Dict[str, Any]], rules: Dict[Tuple[str, ...], str]
) -> Tuple[List[Dict[str, Any]], bool]:
    i = 0
    flg = False
    while i < len(qir) - 1:
        j = _find_next(qir, i)

        if j is not None:
            if (qir[i]["gatef"].n, qir[j]["gatef"].n) in rules:
                nn = rules[(qir[i]["gatef"].n, qir[j]["gatef"].n)]
                if nn == "i":
                    del qir[i]
                    del qir[j - 1]
                    # qir[i] = None
                    # qir[j] = None
                else:
                    param = {}
                    if nn.startswith("r") or nn.startswith("cr"):
                        param = {"theta": _get_theta(qir[i]) + _get_theta(qir[j])}
                    qir[i] = {
                        "gatef": getattr(gates, nn),
                        "name": nn,
                        "mpo": False,
                        "split": False,
                        "parameters": param,
                        "index": qir[i]["index"],
                    }
                    del qir[j]
                    # qir[j] = None
                flg = True
                # return qir, flg
            elif (
                qir[i]["gatef"].n == qir[j]["gatef"].n + "d"
                or qir[i]["gatef"].n + "d" == qir[j]["gatef"].n
            ):
                del qir[i]
                del qir[j - 1]
                # qir[i] = None
                # qir[j] = None
                flg = True
                # return qir, True
        i += 1
    return qir, flg


def merge(
    circuit: Union[AbstractCircuit, List[Dict[str, Any]]],
    rules: Optional[Dict[Tuple[str, ...], str]] = None,
    **kws: Any
) -> Any:
    merge_rules = copy(default_merge_rules)
    if rules is not None:
        merge_rules.update(rules)  # type: ignore
    if isinstance(circuit, list):
        qir = circuit
        output = "qir"
    else:
        qir = circuit.to_qir()
        output = "tc"
    flg = True
    while flg:
        qir, flg = _merge(qir, merge_rules)  # type: ignore
    if output in ["qir"]:
        return qir
    elif output in ["tc", "circuit"]:
        c: Any = type(circuit).from_qir(qir, circuit.circuit_param)  # type: ignore
        return c


def simple_compile(
    circuit: Any,
    info: Optional[Dict[str, Any]] = None,
    output: str = "tc",
    compiled_options: Optional[Dict[str, Any]] = None,
) -> Any:
    if compiled_options is None:
        compiled_options = {}
    len0 = len(circuit.to_qir())
    for d in circuit._extra_qir:
        if d["pos"] < len0 - 1:
            raise ValueError(
                "TC's simple compiler doesn't support measurement/reset instructions \
                in the middle of the circuit"
            )

    c = replace_r(circuit, **compiled_options)
    c = replace_u(circuit, **compiled_options)
    qir = c.to_qir()
    len0 = len(qir)
    qir = merge(qir, **compiled_options)
    qir = prune(qir, **compiled_options)
    len1 = len(qir)
    while len1 != len0:
        len0 = len1
        qir = merge(qir, **compiled_options)
        if len(qir) == len0:
            break
        qir = prune(qir, **compiled_options)
        len1 = len(qir)

    c = type(circuit).from_qir(qir, circuit.circuit_param)

    for d in circuit._extra_qir:
        d["pos"] = len1
        c._extra_qir.append(d)
    return (c, info)



"""
dict related functionalities
"""
from typing import Any, Dict, Optional, Sequence

import numpy as np


Tensor = Any
ct = Dict[str, int]


def reverse_count(count: ct) -> ct:
    ncount = {}
    for k, v in count.items():
        ncount[k[::-1]] = v
    return ncount


def sort_count(count: ct) -> ct:
    return {k: v for k, v in sorted(count.items(), key=lambda item: -item[1])}


def normalized_count(count: ct) -> Dict[str, float]:
    shots = sum([v for k, v in count.items()])
    return {k: v / shots for k, v in count.items()}


def marginal_count(count: ct, keep_list: Sequence[int]) -> ct:
    import qiskit

    count = reverse_count(count)
    ncount = qiskit.result.utils.marginal_distribution(count, keep_list)
    return reverse_count(ncount)


def count2vec(count: ct, normalization: bool = True) -> Tensor:
    nqubit = len(list(count.keys())[0])
    probability = [0] * 2**nqubit
    shots = sum([v for k, v in count.items()])
    for k, v in count.items():
        if normalization is True:
            v /= shots  # type: ignore
        probability[int(k, 2)] = v
    return np.array(probability)


def vec2count(vec: Tensor, prune: bool = False) -> ct:
    from ..quantum import count_vector2dict

    if isinstance(vec, list):
        vec = np.array(vec)
    n = int(np.log(vec.shape[0]) / np.log(2) + 1e-9)
    c = count_vector2dict(vec, n, key="bin")
    if prune is True:
        nc = c.copy()
        for k, v in c.items():
            if np.abs(v) < 1e-8:
                del nc[k]
        return nc
    return c


def kl_divergence(c1: ct, c2: ct) -> float:
    eps = 1e-4  # typical value for inverse of the total shots
    c1 = normalized_count(c1)  # type: ignore
    c2 = normalized_count(c2)  # type: ignore
    kl = 0
    for k, v in c1.items():
        kl += v * (np.log(v) - np.log(c2.get(k, eps)))
    return kl


def expectation(
    count: ct, z: Optional[Sequence[int]] = None, diagonal_op: Optional[Tensor] = None
) -> float:
    """
    compute diagonal operator expectation value from bit string count dictionary

    :param count: count dict for bitstring histogram
    :type count: ct
    :param z: if defaults as None, then ``diagonal_op`` must be set
        a list of qubit that we measure Z op on
    :type z: Optional[Sequence[int]]
    :param diagoal_op: shape [n, 2], explicitly indicate the diagonal op on each qubit
        eg. [1, -1] for z [1, 1] for I, etc.
    :type diagoal_op: Tensor
    :return: the expectation value
    :rtype: float
    """
    if z is None and diagonal_op is None:
        raise ValueError("One of `z` and `diagonal_op` must be set")
    n = len(list(count.keys())[0])
    if z is not None:
        diagonal_op = [[1, -1] if i in z else [1, 1] for i in range(n)]
    r = 0
    shots = 0
    for k, v in count.items():
        cr = 1.0
        for i in range(n):
            cr *= diagonal_op[i][int(k[i])]  # type: ignore
        r += cr * v  # type: ignore
        shots += v
    return r / shots


def plot_histogram(data: Any, **kws: Any) -> Any:
    """
    See ``qiskit.visualization.plot_histogram``:
    https://qiskit.org/documentation/stubs/qiskit.visualization.plot_histogram.html

    interesting kw options include: ``number_to_keep`` (int)

    :param data: _description_
    :type data: Any
    :return: _description_
    :rtype: Any
    """
    from qiskit.visualization import plot_histogram

    return plot_histogram(data, **kws)



# pylint: disable=invalid-name

import sys
import os
from functools import partial
import numpy as np
import tensorflow as tf
import pytest
from pytest_lazyfixture import lazy_fixture as lf

thisfile = os.path.abspath(__file__)
modulepath = os.path.dirname(os.path.dirname(thisfile))

sys.path.insert(0, modulepath)
import tensorcircuit as tc
from tensorcircuit import experimental
from tensorcircuit.quantum import PauliString2COO, PauliStringSum2COO
from tensorcircuit.applications.vqes import construct_matrix_v2

i, x, y, z = [t.tensor for t in tc.gates.pauli_gates]

# note i is in use!

check_pairs = [
    ([0, 0], np.eye(4)),
    ([0, 1], np.kron(i, x)),
    ([2, 1], np.kron(y, x)),
    ([3, 1], np.kron(z, x)),
    ([3, 2, 2, 0], np.kron(np.kron(np.kron(z, y), y), i)),
    ([0, 1, 1, 1], np.kron(np.kron(np.kron(i, x), x), x)),
]


def test_about():
    print(tc.about())


def test_ps2coo(tfb):
    for l, a in check_pairs:
        r1 = PauliString2COO(tf.constant(l, dtype=tf.int64))
        np.testing.assert_allclose(tc.backend.to_dense(r1), a, atol=1e-5)


def test_pss2coo(tfb):
    l = [t[0] for t in check_pairs[:4]]
    a = sum([t[1] for t in check_pairs[:4]])
    r1 = PauliStringSum2COO(tf.constant(l, dtype=tf.int64))
    np.testing.assert_allclose(tc.backend.to_dense(r1), a, atol=1e-5)
    l = [t[0] for t in check_pairs[4:]]
    a = sum([t[1] for t in check_pairs[4:]])
    r1 = PauliStringSum2COO(tf.constant(l, dtype=tf.int64), weight=[0.5, 1])
    a = check_pairs[4][1] * 0.5 + check_pairs[5][1] * 1.0
    np.testing.assert_allclose(tc.backend.to_dense(r1), a, atol=1e-5)


def test_sparse(benchmark, tfb):
    def sparse(h):
        return PauliStringSum2COO(h)

    h = [[1 for _ in range(12)], [2 for _ in range(12)]]
    h = tf.constant(h, dtype=tf.int64)
    sparse(h)
    benchmark(sparse, h)


def test_dense(benchmark, tfb):
    def dense(h):
        return construct_matrix_v2(h, dtype=tf.complex64)

    h = [[1 for _ in range(12)], [2 for _ in range(12)]]
    h = [[1.0] + hi for hi in h]
    dense(h)
    benchmark(dense, h)


@pytest.mark.parametrize("backend", [lf("tfb"), lf("jaxb")])
def test_adaptive_vmap(backend):
    def f(x):
        return x**2

    x = tc.backend.ones([30, 2])

    vf = experimental.adaptive_vmap(f, chunk_size=6)
    np.testing.assert_allclose(vf(x), tc.backend.ones([30, 2]), atol=1e-5)

    vf2 = experimental.adaptive_vmap(f, chunk_size=7)
    np.testing.assert_allclose(vf2(x), tc.backend.ones([30, 2]), atol=1e-5)

    def f2(x):
        return tc.backend.sum(x)

    vf3 = experimental.adaptive_vmap(f2, chunk_size=7)
    np.testing.assert_allclose(vf3(x), 2 * tc.backend.ones([30]), atol=1e-5)

    vf3_jit = tc.backend.jit(vf3)
    np.testing.assert_allclose(vf3_jit(x), 2 * tc.backend.ones([30]), atol=1e-5)


@pytest.mark.parametrize("backend", [lf("tfb"), lf("jaxb")])
def test_adaptive_vmap_mul_io(backend):
    def f(x, y, a):
        return x + y + a

    vf = experimental.adaptive_vmap(f, chunk_size=6, vectorized_argnums=(0, 1))
    x = tc.backend.ones([30, 2])
    a = tc.backend.ones([2])
    # jax vmap has some weird behavior in terms of keyword arguments...
    # TODO(@refraction-ray): further investigate jax vmap behavior with kwargs
    np.testing.assert_allclose(vf(x, x, a), 3 * tc.backend.ones([30, 2]), atol=1e-5)


@pytest.mark.parametrize("backend", [lf("tfb"), lf("jaxb")])
def test_qng(backend):
    n = 6

    def f(params):
        params = tc.backend.reshape(params, [4, n])
        c = tc.Circuit(n)
        c = tc.templates.blocks.example_block(c, params)
        return c.state()

    params = tc.backend.ones([4 * n])
    fim = experimental.qng(f)(params)
    assert tc.backend.shape_tuple(fim) == (4 * n, 4 * n)
    print(experimental.dynamics_matrix(f)(params))


@pytest.mark.parametrize("backend", [lf("tfb"), lf("jaxb")])
def test_dynamic_rhs(backend):
    h1 = tc.array_to_tensor(tc.gates._z_matrix)

    def f(param):
        c = tc.Circuit(1)
        c.rx(0, theta=param)
        return c.state()

    rhsf = experimental.dynamics_rhs(f, h1)
    np.testing.assert_allclose(rhsf(tc.backend.ones([])), -np.sin(1.0) / 2, atol=1e-5)

    h2 = tc.backend.coo_sparse_matrix(
        indices=tc.array_to_tensor(np.array([[0, 0], [1, 1]]), dtype="int64"),
        values=tc.array_to_tensor(np.array([1, -1])),
        shape=[2, 2],
    )

    rhsf = experimental.dynamics_rhs(f, h2)
    np.testing.assert_allclose(rhsf(tc.backend.ones([])), -np.sin(1.0) / 2, atol=1e-5)


@pytest.mark.parametrize("backend", ["tensorflow", "jax"])
def test_two_qng_approaches(backend):
    n = 6
    nlayers = 2
    with tc.runtime_backend(backend) as K:
        with tc.runtime_dtype("complex128"):

            def state(params):
                params = K.reshape(params, [2 * nlayers, n])
                c = tc.Circuit(n)
                c = tc.templates.blocks.example_block(c, params, nlayers=nlayers)
                return c.state()

            params = K.ones([2 * nlayers * n])
            params = K.cast(params, "float32")
            n1 = experimental.qng(state)(params)
            n2 = experimental.qng2(state)(params)
            np.testing.assert_allclose(n1, n2, atol=1e-7)


def test_arg_alias():
    @partial(tc.utils.arg_alias, alias_dict={"theta": ["alpha", "gamma"]})
    def f(theta: float, beta: float) -> float:
        """
        f doc

        :param theta: theta angle
        :type theta: float
        :param beta: beta angle
        :type beta: float
        :return: sum angle
        :rtype: float
        """
        return theta + beta

    np.testing.assert_allclose(f(beta=0.2, alpha=0.1), 0.3, atol=1e-5)
    print(f.__doc__)
    assert len(f.__doc__.strip().split("\n")) == 12


def test_finite_difference_tf(tfb):
    def f(param1, param2):
        n = 4
        c = tc.Circuit(n)
        for i in range(n):
            c.rx(i, theta=param1[i])
        for i in range(n - 1):
            c.cx(i, i + 1)
        for i in range(n - 1):
            c.rzz(i, i + 1, theta=param2[i])
        r = [c.expectation_ps(z=[i]) for i in range(n)]
        return tc.backend.stack(r)

    def fsum(param1, param2):
        return tc.backend.mean(f(param1, param2))

    p1 = tf.ones([4])
    p2 = tf.ones([3])
    g1, g2 = tc.backend.value_and_grad(fsum)(p1, p2)

    f1 = experimental.finite_difference_differentiator(
        f, argnums=(0, 1), shifts=(np.pi / 2, 2)
    )

    def fsum1(param1, param2):
        return tc.backend.mean(f1(param1, param2))

    g3, g4 = tc.backend.value_and_grad(fsum1)(p1, p2)

    np.testing.assert_allclose(g1, g3, atol=1e-5)
    np.testing.assert_allclose(g2, g4, atol=1e-5)


def test_evol(jaxb):
    def h_square(t, b):
        return (tc.backend.sign(t - 1.0) + 1) / 2 * b * tc.gates.x().tensor

    c = tc.Circuit(3)
    c.x(0)
    c.cx(0, 1)
    c.h(2)
    c = experimental.evol_local(
        c, [1], h_square, 2.0, tc.backend.convert_to_tensor(0.2)
    )
    c.rx(1, theta=np.pi - 0.4)
    np.testing.assert_allclose(c.expectation_ps(z=[1]), 1.0, atol=1e-5)

    ixi = tc.quantum.PauliStringSum2COO([[0, 1, 0]])

    def h_square_sparse(t, b):
        return (tc.backend.sign(t - 1.0) + 1) / 2 * b * ixi

    c = tc.Circuit(3)
    c.x(0)
    c.cx(0, 1)
    c.h(2)
    c = experimental.evol_global(
        c, h_square_sparse, 2.0, tc.backend.convert_to_tensor(0.2)
    )
    c.rx(1, theta=np.pi - 0.4)
    np.testing.assert_allclose(c.expectation_ps(z=[1]), 1.0, atol=1e-5)



"""
Interfaces bridging different backends
"""

from . import tensortrans
from .tensortrans import (
    which_backend,
    numpy_args_to_backend,
    general_args_to_numpy,
    general_args_to_backend,
    args_to_tensor,
)
from .numpy import numpy_interface, np_interface
from .scipy import scipy_interface, scipy_optimize_interface
from .torch import torch_interface, pytorch_interface, torch_interface_kws
from .tensorflow import tensorflow_interface, tf_interface


# TODO(@refraction-ray): jax interface using puer_callback and custom_vjp
